/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 * <p>
 * http://www.apache.org/licenses/LICENSE-2.0
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.droolstest;


import org.drools.compiler.kproject.ReleaseIdImpl;
import org.drools.core.io.impl.ReaderResource;
import org.junit.Assert;
import org.junit.Test;
import org.kie.api.KieServices;
import org.kie.api.builder.KieFileSystem;
import org.kie.api.builder.Message;
import org.kie.api.builder.ReleaseId;
import org.kie.api.builder.Results;
import org.kie.api.runtime.KieContainer;
import org.kie.api.runtime.KieSession;

import java.io.IOException;
import java.io.StringReader;
import java.net.URISyntaxException;
import java.time.Instant;
import java.util.*;

public class TestKieContainerUpdate {

	private static final String GROUP_ID = "org.drools.update.test";
	private static final String ARTIFACT_ID = "rules";
	private static final String VERSION_V1 = "1.0.0";
	private static final String VERSION_V2 = "2.0.0";
	private static final String VERSION_V3 = "3.0.0";

	private final static Map<String, String> rules = new HashMap<String, String>() {{
		put("base-rule", "import com.droolstest.TestKieContainerUpdate.TestFact\n" +
				"\n" +
				"dialect  \"mvel\"\n" +
				"\n" +
				"global java.util.List list\n" +
				"\n" +
				"declare TestFact\n" +
				"    @role( event )\n" +
				"    @timestamp ( instant )\n" +
				"    @expires( 1d )\n" +
				"end\n" +
				"\n" +
				"rule \"proxy-rule\"\n" +
				"    when\n" +
				"        $cumulative: Long(longValue>0) from accumulate(\n" +
				"            $g: Long() from entry-point DEFAULT,\n" +
				"            init(long total = 0),\n" +
				"            action(total += $g),\n" +
				"            result(total)\n" +
				"        )\n" +
				"    then\n" +
				"        list.add($cumulative);\n" +
				"end\n");
		put("broken-rule", "import com.droolstest.TestKieContainerUpdate.TestFact\n" +
				"\n" +
				"dialect  \"mvel\"\n" +
				"global java.util.List list\n" +
				"\n" +
				"declare TestFact\n" +
				"    @role( event )\n" +
				"    @timestamp ( instant.toFail() )\n" +
				"    @expires( 1d )\n" +
				"end\n" +
				"\n" +
				"rule \"broken-rule\"\n" +
				"    when\n" +
				"        $g: TestFact($ppid: ppid) over window:length(1)\n" +
				"        $cumulative: Long(longValue>0) from accumulate(\n" +
				"            TestFact(ppid == $ppid, $amount: amount),\n" +
				"            init(long total = 0),\n" +
				"            action(total += $amount),\n" +
				"            result(total)\n" +
				"        )\n" +
				"    then\n" +
				"        System.out.println(\"Version1:-> Amount [\"+$g.getAmount()+\"] - Cumulative [\"+$cumulative+\"]\");\n" +
				"        channels[\"additions\"].send($cumulative);\n" +
				"end\n");
		put("fixed-rule", "import com.droolstest.TestKieContainerUpdate.TestFact\n" +
				"\n" +
				"dialect  \"mvel\"\n" +
				"global java.util.List list\n" +
				"\n" +
				"declare TestFact\n" +
				"    @role( event )\n" +
				"    @timestamp ( instant )\n" +
				"    @expires( 1d )\n" +
				"end\n" +
				"\n" +
				"rule \"broken-rule\"\n" +
				"    when\n" +
				"        $g: TestFact($ppid: ppid) over window:length(1)\n" +
				"        $cumulative: Long(longValue>0) from accumulate(\n" +
				"            TestFact(ppid == $ppid, $amount: amount),\n" +
				"            init(long total = 0),\n" +
				"            action(total += $amount),\n" +
				"            result(total)\n" +
				"        )\n" +
				"    then\n" +
				"        System.out.println(\"Version1:-> Amount [\"+$g.getAmount()+\"] - Cumulative [\"+$cumulative+\"]\");\n" +
				"        channels[\"additions\"].send($cumulative);\n" +
				"end\n");
	}};

	private final static String kmoduleTemplate = "<kmodule xmlns=\"http://jboss.org/kie/6.0.0/kmodule\">\n" +
			"    <kbase name=\"kbase-test\" default=\"true\" equalsBehavior=\"equality\" eventProcessingMode=\"stream\">\n" +
			"        <ksession name=\"ksession-test\" default=\"true\" type=\"stateful\" clockType=\"pseudo\"/>\n" +
			"    </kbase>\n" +
			"</kmodule>";

	private final static String pomTemplate = "<project xmlns=\"http://maven.apache.org/POM/4.0.0\" xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" xsi:schemaLocation=\"http://maven.apache.org/POM/4.0.0 http://maven.apache.org/xsd/maven-4.0.0.xsd\">\n" +
			"  <modelVersion>4.0.0</modelVersion>\n" +
			"  <groupId>%s</groupId>\n" +
			"  <artifactId>%s</artifactId>\n" +
			"  <version>%s</version>\n" +
			"</project>";

	private static Results buildRelease(ReleaseId releaseId, String... resources) {

		String pomFile = String.format(pomTemplate, releaseId.getGroupId(), releaseId.getArtifactId(), releaseId.getVersion());
		ReaderResource pomResource = new ReaderResource(new StringReader(pomFile));
		ReaderResource kmodule = new ReaderResource(new StringReader(kmoduleTemplate));

		KieServices kieServices = KieServices.Factory.get();

		KieFileSystem kfs = kieServices.newKieFileSystem();

		kfs.write("pom.xml", pomResource);

		kfs.write("src/main/resources/META-INF/kmodule.xml", kmodule);

		for (String res : resources) {
			ReaderResource ruleResource = new ReaderResource(new StringReader(rules.get(res)), "utf-8");
			System.out.println("loading rules in kbase " + res + " " + ruleResource);
			kfs.write("src/main/resources/" + res + ".drl", ruleResource);
		}

		Results results = kieServices.newKieBuilder(kfs).buildAll().getResults();
		logResults(results);

		return results;
	}

	@Test
	public void testUpdateRollback() throws IOException, URISyntaxException {

		KieServices kieServices = KieServices.Factory.get();

		ReleaseIdImpl releaseIdV1 = new ReleaseIdImpl(GROUP_ID, ARTIFACT_ID, VERSION_V1);
		ReleaseIdImpl releaseIdV2 = new ReleaseIdImpl(GROUP_ID, ARTIFACT_ID, VERSION_V2);


		Results results = buildRelease(releaseIdV1, "base-rule");
		Assert.assertTrue("Expecting no errors in 1.0.0", !results.hasMessages(Message.Level.ERROR));
		results = buildRelease(releaseIdV2, "base-rule", "broken-rule");
		Assert.assertTrue("Expecting no errors in 2.0.0", !results.hasMessages(Message.Level.ERROR));

		KieContainer kieContainer = kieServices.newKieContainer(releaseIdV1);
		System.out.println("Added kjar 1.0.0 with a good rule");


		KieSession kieSession = kieContainer.newKieSession();
		ArrayList<Object> globalList = new ArrayList<>();
		kieSession.setGlobal("list", globalList);


		kieSession.insert(1L);
		kieSession.fireAllRules();


		kieSession.insert(2L);
		kieSession.fireAllRules();

		System.out.println("Updating to kjar 2.0.0 with previous rule and a broken one");

		results = kieContainer.updateToVersion(releaseIdV2);
		System.out.println("Expecting error for the broken rule");
		Assert.assertTrue("Expecting errors in 2.0.0", results.hasMessages(Message.Level.ERROR));

		System.out.println("Upgrading to the previous version as rollback.");
		try {
			System.out.println("Expecting nullPointerException trying to remove the rule");
			results = kieContainer.updateToVersion(releaseIdV1);
		} catch (Exception e) {
			System.out.println("Found exception rollbacking");
			e.printStackTrace();
			Assert.assertTrue("There is a nullPointerException", e instanceof NullPointerException);
		}

		kieSession.dispose();
	}

	@Test
	public void itShouldFailKJarBuild() throws IOException, URISyntaxException {

		final ReleaseIdImpl releaseIdV1 = new ReleaseIdImpl(GROUP_ID, ARTIFACT_ID, VERSION_V1);
		final ReleaseIdImpl releaseIdV2 = new ReleaseIdImpl(GROUP_ID, ARTIFACT_ID, VERSION_V2);

		System.out.println("Creating kjar 1.0.0");
		Results results = buildRelease(releaseIdV1, "base-rule");

		Assert.assertTrue("Expecting no errors in 1.0.0 KJar", !results.hasMessages(Message.Level.ERROR));

		System.out.println("Creating kjar 2.0.0 with previous rule and a broken one");

		results = buildRelease(releaseIdV2, "broken-rule");

		Assert.assertTrue("Expecting errors in 2.0.0 KJar", results.hasMessages(Message.Level.ERROR));

	}

	@Test
	public void testUpdateFixingRuleInNewKJAR() throws IOException, URISyntaxException {
		ArrayList<Object> globalList = new ArrayList<>();
		KieServices kieServices = KieServices.Factory.get();

		ReleaseIdImpl releaseIdV1 = new ReleaseIdImpl(GROUP_ID, ARTIFACT_ID, VERSION_V1);
		ReleaseIdImpl releaseIdV2 = new ReleaseIdImpl(GROUP_ID, ARTIFACT_ID, VERSION_V2);
		ReleaseIdImpl releaseIdV3 = new ReleaseIdImpl(GROUP_ID, ARTIFACT_ID, VERSION_V3);

		Results results = buildRelease(releaseIdV1, "base-rule");
		Assert.assertTrue("Expecting no errors in 1.0.0 KJar", !results.hasMessages(Message.Level.ERROR));
		results = buildRelease(releaseIdV2, "base-rule", "broken-rule");
		Assert.assertTrue("Expecting no errors in 2.0.0 KJar", results.hasMessages(Message.Level.ERROR));
		buildRelease(releaseIdV3, "base-rule", "fixed-rule");
		Assert.assertTrue("Expecting no errors in 3.0.0 KJar", results.hasMessages(Message.Level.ERROR));

		//Creating default kjar v1.0.0
		KieContainer kieContainer = kieServices.newKieContainer(releaseIdV1);
		System.out.println("Added kjar 1.0.0 with a good rule");


		KieSession kieSession = kieContainer.newKieSession();
		kieSession.setGlobal("list", globalList);

		Assert.assertEquals(0, globalList.size());


		System.out.println("Updating to kjar 2.0.0 with previous rule and a broken one");

		results = kieContainer.updateToVersion(releaseIdV2);
		Assert.assertTrue("Expecting errors in 2.0.0 upgrade", results.hasMessages(Message.Level.ERROR));

		System.out.println("Upgrading to kjar 3.0.0 with the broken rule fixed");
		try {
			System.out.println("Expecting nullPointerException trying to update the rule");
			results = kieContainer.updateToVersion(releaseIdV3);
		} catch (Exception e) {
			System.out.println("Found exception upgrading to 3.0.0");
			e.printStackTrace();
			Assert.assertTrue("There is a nullPointerException", e instanceof NullPointerException);
		}

		kieSession.dispose();
	}

	public static void logResults(Results results) {
		for (Message result : results.getMessages()) {
			switch (result.getLevel()) {
				case ERROR:
					System.out.println("error " + result.toString());
					break;
				case WARNING:
					System.out.println("warn " + result.toString());
					break;
				case INFO:
					System.out.println("info " + result.toString());
					break;
				default:
					System.out.println("warn " + result.toString());
			}
		}
	}

	public static class TestFact {

		private final long ppid;
		private long amount;
		private final Date instant;

		public TestFact(long ppid, long amount, Date instant) {
			this.ppid = ppid;
			this.amount = amount;
			this.instant = instant;
		}

		public long getPpid() {
			return ppid;
		}

		public long getAmount() {
			return amount;
		}

		public Instant getInstant() {
			return instant.toInstant();
		}

		@Override
		public boolean equals(Object o) {
			if (this == o) return true;
			if (!(o instanceof TestFact)) return false;
			TestFact testFact = (TestFact) o;
			return ppid == testFact.ppid &&
					amount == testFact.amount &&
					Objects.equals(instant, testFact.instant);
		}

		@Override
		public int hashCode() {
			return Objects.hash(ppid, amount, instant);
		}

		@Override
		public String toString() {
			return "TestFact{" +
					"ppid=" + ppid +
					", amount=" + amount +
					", instant=" + instant +
					'}';
		}

		public void setAmount(long newAmount) {
			this.amount = newAmount;
		}
	}
}
